/*
 * Global Memory and Threading (GMT)
 *
 * Copyright © 2018, Battelle Memorial Institute
 * All rights reserved.
 *
 * Battelle Memorial Institute (hereinafter Battelle) hereby grants permission to
 * any person or entity lawfully obtaining a copy of this software and associated
 * documentation files (hereinafter “the Software”) to redistribute and use the
 * Software in source and binary forms, with or without modification.  Such
 * person or entity may use, copy, modify, merge, publish, distribute,
 * sublicense, and/or sell copies of the Software, and may permit others to do
 * so, subject to the following conditions:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name `Battelle Memorial Institute` or `Battelle` may be used in
 *    any form whatsoever without the express written consent of `Battelle`.
 *  
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL `BATTELLE` OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef __DEBUG_H__
#define __DEBUG_H__

#include <stdlib.h>
#include <stdint.h>
#include <unistd.h>
#include <execinfo.h>

#include "gmt/config.h"
#include "gmt/utils.h"

#define _unused(x) ((void)x)
#define NO_INSTRUMENT  __attribute__((no_instrument_function))

#define DEBUG1(x) x
#define DEBUG0(x)

#if DISABLE_INLINE
#define INLINE
#else
#define INLINE static inline
#endif

extern uint32_t node_id;
extern void *pt_stacks;
extern uint64_t pt_stacks_size;
extern void *ut_stacks;
extern uint64_t ut_stacks_size;
extern char *prog_name;
void debug_init();
uint32_t get_thread_id() NO_INSTRUMENT;

#if defined(__cplusplus)
extern "C" {
#endif

    void _debug(const char *func, const char *file, int line, char *fmt, ...) NO_INSTRUMENT;
    void btrace() NO_INSTRUMENT;

#if defined(__cplusplus)
}
#endif

#define _DEBUG(format, ...) _debug(__func__, __FILE__ , __LINE__,\
(char*)format, ## __VA_ARGS__);

#if ENABLE_ASSERTS
#define _assert(x)  {\
if (!(x))  { \
    _DEBUG("Assert error: (%s)\n", #x); \
    btrace();\
    _exit(EXIT_FAILURE);\
}}
#else
#define _assert(x) { _unused((x)); }
#endif

#define _check(x) {\
    if(!(x)) { \
        _DEBUG("Check failed (%s)\n", #x); \
        _exit(EXIT_FAILURE);\
    }\
}

#define ERRORMSG(format, ...) {   \
  _debug(__func__, __FILE__ , __LINE__, (char*)format, ## __VA_ARGS__);\
  btrace();\
  _exit(EXIT_FAILURE);}
#endif
